/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "segdb.hh"
#include "pool_mgr.hh"
#include "buffer.hh"
#include <stdexcept>
#include <iostream>

using namespace sends;
using namespace std;
using namespace thread;

//======================================  Segment database constructor.
seg_db::seg_db(seg_index nSeg, chan_index nChans) 
    : mNSegs(nSeg), mAgeQ(nSeg), mAgeLinks(nSeg),  mChanBase(nChans,nSeg), 
      mChanLinks(nSeg), mSegVect(nSeg)
{
    semlock mlock(mAgeQ.ref_mutex(), semlock::wrlock);
    for (seg_index i=0; i<nSeg; ++i) {
	mAgeQ.insert(i, end(), mAgeLinks);
    }
}

//======================================  Segment database destructor.
seg_db::~seg_db(void) {
    clear();
}

//======================================  Add a new segment that doesn't
//                                        overlap existing segments.
seg_index
seg_db::add_segment(pool_mgr& pmgr, chan_index chan, double rate, 
		    data_type dtyp, gps_type start, gps_type mind) {

    //----------------------------------  Get a free segment from the end
    //                                    of the age queue. Fill in its id.
    gps_type stop = start + mind;
    int lmin = int(rate * dtyp.size() * mind + 0.5);
    seg_index inx = get_segment(pmgr, pmgr.get_buffer_class(lmin));
    mSegVect[inx].set_segment(chan, rate, dtyp, start);
    if (mSegVect[inx].time_left() < mind) {
	mSegVect[inx].clear();
	free(inx);
	cerr << "Buffer for channel: " << chan 
	     << " is not long enough to hold " 
	     << mind << "s of data." << endl;
	return end();
    }

    //----------------------------------  See if it fits in the list.
    try	{
	mChanBase[chan].write_lock();

	//------------------------------  Find the previous segment
	seg_index iseg = last_leq_int(chan, start);
	gps_type  pMax = mSegVect[iseg].getMaxGPS();

	//------------------------------  Test if a segment for here exists
	//    Note that this is no good if the the minimum segment won't fit in.
	if (!is_null(iseg) && start <= mSegVect[iseg].getEndGPS() &&
	    stop <= pMax) {
	    mChanBase[chan].unlock();
	    mSegVect[inx].clear();
	    free(inx);
	    return iseg;
	}

	//------------------------------  Truncate previous if it overlaps new
	if (!is_null(iseg) && pMax > start) {
	    mSegVect[iseg].cap(start);
	}

	//------------------------------  Truncate new segment if it overlaps
	seg_index jnx = iseg;
	if (is_null(iseg)) jnx = mChanBase[chan].first();
	else               jnx = mChanBase[chan].next(iseg, mChanLinks);
	if (!is_null(jnx)) {
	    gps_type  nxtStart = mSegVect[jnx].getStartGPS();
	    if (mSegVect[inx].getMaxGPS() > nxtStart) {
		mSegVect[inx].cap(nxtStart);
	    }
	}

	//------------------------------  Insert segment in the channel list
	seg_index  ipos = first_gt_int(chan, mSegVect[inx].getStartGPS());
	mChanBase[chan].insert(inx, ipos, mChanLinks);
	mChanBase[chan].unlock();
    } catch (exception& e) {
	mChanBase[chan].unlock();
	mSegVect[inx].clear();
	free(inx);
	throw;
    }

    //----------------------------------  Insert the segment in the Age queue
    make_new(inx);
    return inx;
}

//======================================  Check all segments in the age queue
void
seg_db::check(void) const {
    semlock mlock(mAgeQ.ref_mutex(), semlock::rdlock);
    seg_index age = 0;
    seg_index later = mNSegs;
    for (seg_index i=mAgeQ.last(); !is_null(i) && age<mNSegs; 
	 i=mAgeLinks.prev(i)) {
	//------------------------------  Check the age queue forward link
	if (mAgeLinks.next(i) != later) {
	    cerr << "seg_db::check: Incorrect forward link, expected: " 
		 << later << ", found: " << mAgeLinks.next(i) << endl;
	}
	later = i;

	//------------------------------  Is it in a channel list?
	chan_index chan = mSegVect[i].chan();
	if (chan_valid(chan)) {
	    semlock mlock(mChanBase[chan].ref_mutex(), semlock::rdlock);

	    //--------------------------  Check the back link
	    seg_index cinx = mChanLinks.prev(i);
	    if (is_null(cinx)) {
		if (mChanBase[chan].first() != i) {
		    cerr << "seg_db::check: Channel: " << chan << " segment: " 
			 << i << " previous is null." << endl;
		}
	    }
	    else if (mChanLinks.next(cinx) != i) {
		cerr << "seg_db::check: Channel: " << chan << " segment: " 
		     << i << " previous doesn't link to current." << endl;
	    }

	    //-------------------------  Check the forward link
	    cinx = mChanLinks.next(i);
	    if (is_null(cinx)) {
		if (mChanBase[chan].last() != i) {
		    cerr << "seg_db::check: Channel: " << chan << " segment: " 
			 << i << " next is null." << endl;
		}
	    }
	    else if (mChanLinks.prev(cinx) != i) {
		cerr << "seg_db::check: Channel: " << chan << " segment: " 
		     << i << " next doesn't link to current." << endl;
	    }
	}
	age++;
    }
}

//======================================  Remove all segments from the database
void
seg_db::clear(void) {
    cout << "Clear all segments..." << endl;
    check();
    semlock mlock(mAgeQ.ref_mutex(), semlock::wrlock);
    for (seg_index i=0; i<end(); ++i) {
	if (mAgeLinks.next(i) != end() || mAgeLinks.prev(i) != end()) {
	    mAgeQ.remove(i, mAgeLinks);
	    seg_index chan = mSegVect[i].chan();
	    if (chan_valid(chan)) {
		semlock mlock(mChanBase[chan].ref_mutex(), semlock::wrlock);
		mChanBase[chan].remove(i, mChanLinks);
		mSegVect[i].release();
	    }
	}
    }
}

//======================================  Dump all segments in the age queue
std::ostream&
seg_db::dump(std::ostream& out) const {
    check();
    out << "seg age clst cnxt ";
    segment::dump_header(out);
    seg_index age = 0;
    mAgeQ.read_lock();
    for (seg_index i=mAgeQ.last(); !is_null(i) && age<mNSegs; 
	 i=mAgeLinks.prev(i)) {
	if (!mSegVect[i].empty() || mSegVect[i].inUse()) {
	    seg_index inx = mChanLinks.prev(i);
	    out << i << " " << age << " ";
	    if (is_null(inx)) out << "  -  ";
	    else              out << inx << " "; 
	    inx = mChanLinks.next(i);
	    if (is_null(inx)) out << "  -  ";
	    else              out << inx << " "; 
	    mSegVect[i].dump(out);
	}
	age++;
    }
    mAgeQ.unlock();
    return out;
}

//======================================  Dump all segments in the age queue
std::ostream&
seg_db::dump_by_channel(std::ostream& out) const {
    chan_index nChan = mChanBase.size();
    for (chan_index j=0; j<nChan; j++) {
	if (!is_null(mChanBase[j].first())) {
	    out << "Active segments buffers for channel: " << j << endl;
	    out << " age ";
	    segment::dump_header(out);
	    for (seg_index i=mChanBase[j].first(); !is_null(i); 
		 i=mChanLinks.next(i)) {
		if (!mSegVect[i].empty() || mSegVect[i].inUse()) {
		    out << get_age(i) << " ";
		    mSegVect[i].dump(out);
		}
	    }
	}
    }
    return out;
}

//======================================  Return a segment to the list
void 
seg_db::free(seg_index iseg) {
    if (chan_valid(mSegVect[iseg].chan())) {
	throw runtime_error("Attempt to free segment with valid channel");
    }
    semlock mlock(mAgeQ.ref_mutex(), semlock::wrlock);
    mAgeQ.insert(iseg, mAgeQ.first(), mAgeLinks);
}

//======================================  Find a segment
seg_index
seg_db::find(chan_index chan, gps_type start, gps_type stop) const {
    semlock mlock(mChanBase[chan].ref_mutex(), semlock::rdlock);

    //----------------------------------  Skip past early segments
    seg_index inx = last_leq_int(chan, start);

    //----------------------------------  Check for overlap - null if not.
    if (is_null(inx) || stop <= mSegVect[inx].getStartGPS()) {
	inx = end();
    }
    return inx;
}

//======================================  Find first segment after gps
seg_index 
seg_db::first_gt_int(chan_index ichn, gps_type gps) const {
    seg_index iseg = mChanBase[ichn].first();
    while (!is_null(iseg) && mSegVect[iseg].getStartGPS() <= gps) {
	iseg = mChanLinks.next(iseg);
    }
    return iseg;
}

seg_index 
seg_db::first_gt(chan_index ichn, gps_type gps) const {
    semlock mlock(mChanBase[ichn].ref_mutex(), semlock::rdlock);
    return first_gt_int(ichn, gps);
}

//======================================  Find the segments age-queue position
seg_index
seg_db::get_age(seg_index iseg) const {
    semlock mlock(mAgeQ.ref_mutex(), semlock::rdlock);
    seg_index age=0;
    for (seg_index i=mAgeQ.last(); age < end() && !is_null(i); 
	 i=mAgeLinks.prev(i)) 
	{
	    if (iseg == i) return age;
	    age++;
	}
    return end();
}

//======================================  Find/allocate buffer for new data.
seg_index
seg_db::get_buffer(pool_mgr& pmgr, chan_index chan, double rate, data_type typ,
		   gps_type start, gps_type stop) {

    //----------------------------------  Find the closest segment before start
    if (stop <= start) return end();
    seg_index inx = last_leq(chan, start);

    //----------------------------------  Find if a segment exists
    //
    //  At this point inx points to the last segment before, or the segment
    //  containing the data. If inx is null, or the data are after the end 
    //  of the buffer, we have to insert a new buffer.
    //
    if (is_null(inx) || 
	mSegVect[inx].getMaxGPS() < stop || 
	mSegVect[inx].getEndGPS() < start) {
	inx = add_segment(pmgr, chan, rate, typ, start, stop-start);
    }

    //----------------------------------  Lock segment for fill, reserve space
    if (!is_null(inx)) {
	mSegVect[inx].use();
	if (!mSegVect[inx].reserve(start, stop)) {
	    mSegVect[inx].unuse();
	    inx = end();
	}
    }
    return inx;
}

//======================================  Get oldest segment.
seg_index 
seg_db::get_segment(pool_mgr& pmgr, int bufclass) {
    seg_index iseg = end();
    buffer* pbuf = pmgr.get_buffer_now(bufclass);
    if (pbuf) {
	iseg = getOldest(-1);
	if (is_null(iseg)) pbuf->free();
	else               mSegVect[iseg].set_buffer(pbuf, bufclass);
    }
    if (is_null(iseg)) iseg = getOldest(bufclass);
    if (is_null(iseg)) {
	dump(cout);
	throw runtime_error("No free segments");
    }
    return iseg;
}

//======================================  Get oldest segment.
seg_index 
seg_db::getOldest(int bufclass) {
    semlock mlock(mAgeQ.ref_mutex(), semlock::wrlock);

    //----------------------------------  Find oldest appropriate segment
    seg_index old = mAgeQ.first();
    while (!is_null(old) && !mSegVect[old].canUse(bufclass)) {
	old = mAgeLinks.next(old);
    }

    //----------------------------------  Remove it from the list(s)
    if (!is_null(old)) {
	mAgeQ.remove(old, mAgeLinks);
	seg_index chan = mSegVect[old].chan();
	if (chan_valid(chan)) {
	    semlock mlock(mChanBase[chan].ref_mutex(), semlock::wrlock);
	    mChanBase[chan].remove(old, mChanLinks);
	    mSegVect[old].clear();
	}
    }
    return old;
}

//======================================  Insert a segment
void 
seg_db::insert(chan_index ichn, seg_index inx) {
    seg_index  ipos = first_gt_int(ichn, mSegVect[inx].getStartGPS());
    mChanBase[ichn].insert(inx, ipos, mChanLinks);
    mAgeQ.insert(inx, end(), mAgeLinks);
}

//======================================  Find latest segment not after gps
seg_index 
seg_db::last_leq_int(chan_index ichn, gps_type gps) const {
    seg_index last = end();
    for (seg_index iseg=mChanBase[ichn].first();
	 !is_null(iseg) && mSegVect[iseg].getStartGPS() <= gps;
	 iseg = mChanLinks.next(iseg)) {
	last = iseg;
    }
    return last;
}

seg_index 
seg_db::last_leq(chan_index ichn, gps_type gps) const {
    semlock mlock(mChanBase[ichn].ref_mutex(), semlock::rdlock);
    return last_leq_int(ichn, gps);
}

//======================================  Move segment to newest
void 
seg_db::make_new(seg_index inx) {
    semlock mlock(mAgeQ.ref_mutex(), semlock::wrlock);    
    mAgeQ.remove(inx, mAgeLinks);
    mAgeQ.insert(inx, end(), mAgeLinks);
}

//======================================  Release (unuse) the specified segment 
void 
seg_db::release(seg_index iseg) {
    if (is_null(iseg)) return;
    segment& segi = mSegVect[iseg];
    chan_index ichn = segi.chan();
    if (!segi.unuse()) {

	//------------------------------  Blow away empty segments
	if (segi.empty()) {
	    semlock mlock(mAgeQ.ref_mutex(), semlock::wrlock);

	    if (ichn == segi.chan() && segi.empty() && !segi.inUse()) {
		mAgeQ.remove(iseg, mAgeLinks);
		if (chan_valid(ichn)) {
		    semlock clck(mChanBase[ichn].ref_mutex(), semlock::wrlock);
		    mChanBase[ichn].remove(iseg, mChanLinks);
		}
		segi.clear();
		mAgeQ.insert(iseg, mAgeQ.first(), mAgeLinks);
	    }
	}

	//------------------------------  Merge with next segment?
	else {
	    semlock chanlock(mChanBase[ichn].ref_mutex());
	    if (ichn != mSegVect[iseg].chan()) return;

	    seg_index jseg = mChanLinks.next(iseg);
	    if (is_null(jseg)) return;
	    if (mSegVect[iseg].getEndGPS() >= mSegVect[jseg].getStartGPS()) {
		cerr << "Segment: " << iseg << " can be merged with Segment: " 
		     << jseg << endl;
		segment::dump_header(cerr);
		mSegVect[iseg].dump(cerr);
		mSegVect[jseg].dump(cerr);
	    }
	}
    }
}

//======================================  Find and use the specified segment
seg_index
seg_db::reserve(chan_index chan, gps_type gps) {
    mChanBase[chan].read_lock();

    //----------------------------------  Find best guess.
    seg_index inx = last_leq_int(chan, gps);
    segment& seg(mSegVect[inx]);

    //----------------------------------  Mark segment "used" if it contains gps
    if (is_null(inx)) {
	mChanBase[chan].unlock();
    } else if (gps >= seg.getStartGPS() && gps < seg.getReserveGPS()) {
	seg.use();
	mChanBase[chan].unlock();
	if (gps >= seg.getEndGPS()) {
	    seg.fill_wait();   // wait for lock to complete
	}
	make_new(inx);
    } else {
	mChanBase[chan].unlock();
	inx = end();
    }
    return inx;
}

//======================================  Set segment list length
void
seg_db::set_channel_count(chan_index nChan) {
    mChanBase.resize(nChan, mNSegs);
}

//======================================  Insert a segment at the specified 
//                                        point in the queue
void 
seg_db::qbase::insert(seg_index iseg, seg_index ib4, linkage& l) {
    testlock();
    if (!l.null(ib4)) {
	l.insert_before(iseg, ib4);
    } else {
	if (!l.null(mLast)) l.insert_after(iseg, mLast);
	mLast  = iseg;
    }
    if (ib4 == mFirst)  mFirst = iseg;
}

//======================================  Remove the specified segment 
//                                        from the queue
void 
seg_db::qbase::remove(seg_index iseg, linkage& l) {
    testlock();
    if (iseg == mFirst) mFirst = l.next(iseg);
    if (iseg == mLast)  mLast  = l.prev(iseg);
    l.unlink(iseg);
}

//======================================  Lock the queue for read
void
seg_db::qbase::read_lock(void) const {
    mRWMutex.readlock();
}

//======================================  Test whether queue is write-locked
void
seg_db::qbase::testlock(void) const {
    //----------------------------------  Test for write-lock
    //   Note that trylock(wrlock) fails if mutex is read-locked or write-
    //   locked, whereas trylock(rdlock) fails only if semaphore is write-
    //   locked. Unlock and throw an exception if trylock succeeds.
    //
    if (mRWMutex.trylock(readwritelock::rdlock)) {
	mRWMutex.unlock();
	throw logic_error("Attempt to modify unlocked segment list");
    }
}

//======================================  Unlock the queue
void
seg_db::qbase::unlock(void) const {
    mRWMutex.unlock();
}

//======================================  Lock the queue for write
void
seg_db::qbase::write_lock(void) {
    mRWMutex.writelock();
}
